#include "cadef.h"
#include "stdio.h"
#ifndef vxWorks
 #define VER_3_14 
#endif /* vxWorks */
#define caDebug 0

static void printChidInfo(chid chid, char *message) {
  if(caDebug) {
    printf("\n%s\n",message);
    printf("pv: %s  type(%d) nelements(%ld) host(%s)",ca_name(chid),
	   ca_field_type(chid),ca_element_count(chid),ca_host_name(chid));
    printf(" read(%d) write(%d) state(%d)\n",
	   ca_read_access(chid),ca_write_access(chid),ca_state(chid));
  }
}

static void exceptionCallback(struct exception_handler_args args) {
  chid	chid = args.chid;
  long	stat = args.stat; /* Channel access status code*/
  const char  *channel;
  static char *noname = "unknown";

  channel = (chid ? ca_name(chid) : noname);
  if(caDebug) {
    if(chid) printChidInfo(chid,"exceptionCallback");
    printf("exceptionCallback stat %s channel %s\n",ca_message(stat),channel);
  }
}

static void connectionCallback(struct connection_handler_args args) {
chid	chid = args.chid;
    printChidInfo(chid,"connectionCallback");
}

static void accessRightsCallback(struct access_rights_handler_args args) {
chid	chid = args.chid;
   printChidInfo(chid,"accessRightsCallback");
}

static void eventCallback(struct event_handler_args eha) {
chid	chid = eha.chid;
static char value[256];
static char name[128];
char *ptr;
int severity=0;
int status=0;
float valAsFloat;
struct dbr_stsack_string * pStruct = (struct dbr_stsack_string *)eha.dbr;
  status=pStruct->status;
  severity=pStruct->severity;
  memset(value,0,256);
  if(eha.status!=ECA_NORMAL) {
    printChidInfo(chid,"eventCallbackBadStatus");
  } else {
    memset(name,0,128);
    strcpy(name,ca_name(eha.chid));
    if( (ptr=strstr(name,".VAL")) != NULL) *ptr=0;

    strcpy(value,pStruct->value);
    valAsFloat = atof(value);

    if(caDebug) printf("Event Callback for '%s'='%f' [st=%d sev=%d]\n",
		       name,valAsFloat,status,severity);
    tineSetAlarm(name,severity,status,valAsFloat);
  }
}

int e2t_add_event(char *recName, int *pSev)
{
  static int first = 1;
  chid mychid;
  evid myevid;
#ifdef VER_3_14 
  if(first ) {
    first=0;
    SEVCHK(ca_context_create(ca_enable_preemptive_callback),"ca_cnt_create");
    SEVCHK(ca_add_exception_event(exceptionCallback,NULL),"ca_add_exception_event");
  }
  SEVCHK(ca_create_channel(recName,connectionCallback,pSev,20,&mychid),
	 "ca_create_channel");
#else /* VER_3.14 */
  SEVCHK(ca_search_and_connect(recName,&mychid,connectionCallback,pSev), 
	 "ca_create_channel");
  if(ca_pend_io(0.5) != ECA_NORMAL) printf("ca_pend failed\n");
#endif
  SEVCHK(ca_replace_access_rights_event(mychid,accessRightsCallback),
	 "ca_replace_access_rights_event");
  SEVCHK(ca_add_masked_array_event(DBR_STSACK_STRING,1,mychid,eventCallback,pSev,
	 (float) 0, (float) 0, (float) 0,&myevid,DBE_ALARM), "ca_add_event");
	return(1);
}
