#include "e2t.h"
#include "alarm.h" 
#include <stdlib.h>

#ifdef WIN32
#include <stddef.h>
#include <stddef.h>
#include "epicsThread.h"
#include "iocsh.h"
#include <registryFunction.h>
#include "epicsExport.h"
#include <subRecord.h>
#endif  /* WIN32 */

Tproperty	*pT = 0;

extern int NGdebug;
extern void haltServer(void);
extern int max_rcvbuf, max_sndbuf;
extern int e2t_add_event(char *recName, int *pSev);

int makeDeviceNameFromChannelName(char *chName,char *devName);

int numProperties = 0;
int offsetComposites = 0;
COMPOSITE *composites = NULL;
int ncomposites = 0;
int debug = 0;

#ifdef WIN32
int main(int argc,char *argv[])
{
    if(argc>=2) {    
        iocsh(argv[1]);
        epicsThreadSleep(.2);
    }

    iocsh(NULL);
    return(0);
}
#endif	/* WIN32 */

/************************************************************************
 *									*
 *	EPICS Specific Part						*
 *									*
 ************************************************************************/

#ifdef vxWorks
extern struct dbBase *pdbbase;
#else
epicsShareExtern struct dbBase *pdbbase;
#endif	/* vxWorks */

int	CA_OK  	       = 0;
int	ERRORnoDevice  = 1;
int	ERRORnoChannel = 2;
int	ERRORnoValue   = 3;

#ifdef WIN32
static const iocshFuncDef tineTestFuncDef = {"tineGetNumRecords", 0};
static void tineTestCallFunc(const iocshArgBuf *args)
{
	tineGetNumRecords();
}

static void tineTest(void)
{
	iocshRegister(&tineTestFuncDef, tineTestCallFunc);
}

epicsExportRegistrar(tineTest);
/************************************************************************/

int zkDebug = 0 ;
epicsExportAddress(int, zkDebug);

typedef long (*processMethod)(subRecord *precord);

static long zkInit(subRecord *precord,processMethod process)
{
  if (zkDebug)
      printf("Record %s called zkInit(%p, %p)\n",
             precord->name, (void*) precord, (void*) process);
  return(0);
}

static long zkProcess(subRecord *precord)
{
  printf("Record %s called zkProcess(%p)\n",precord->name, (void*) precord);
  startE2T();
  return(0);
}

static registryFunctionRef zkRef[] = {
  {"zkInit",(REGISTRYFUNCTION)zkInit},
  {"zkProcess",(REGISTRYFUNCTION)zkProcess}
};

static void zkRegistrar(void)
{
    registryFunctionRefAdd(zkRef,NELEMENTS(zkRef));
}
epicsExportRegistrar(zkRegistrar);

/***************************************************************************/
#endif	/* WIN32 */

long tineGetNumRecords()
{
  DBENTRY     dbentry;
  DBENTRY     *pdbentry=&dbentry;
  long        status = 0;
  int         nrecords = 0;
  int         total = 0;


  dbInitEntry(pdbbase,pdbentry);
  status = dbFirstRecordType(pdbentry);
  if(status) if(debug) printf("No record description\n");

  while(!status) {
    nrecords = dbGetNRecords(pdbentry);
    total += nrecords;
    status = dbNextRecordType(pdbentry);
  }
  dbFinishEntry(pdbentry);
  if(debug) printf("Total records (tineGetNumRecords): %d\n",total);
  return(total);
}

long tineGetRecords(NAME32 *pNames)
{
  DBENTRY     dbentry;
  DBENTRY     *pdbentry=&dbentry;
  long        status = 0;
  short       c = 0;

  dbInitEntry(pdbbase,pdbentry);
  status = dbFirstRecordType(pdbentry);
  if(debug) if(status) printf("No record description\n");

  while(!status) {
    status = dbFirstRecord(pdbentry);
    while(!status) {
      strncpy(pNames[c].name, dbGetRecordName(pdbentry), 32);
      status = dbNextRecord(pdbentry);
      c++;
    }
    status = dbNextRecordType(pdbentry);
  }
  dbFinishEntry(pdbentry);

  if(debug) printf("Total records (tineGetRecords): %d\n", c);
  return (c);
}

long tineGetNumProperties(char *pname)
{
  return(TINE_FREE_PROPERTIES);
}

long tineGetProperties(char *pname, Tproperty *pprop)
{
  DZK(printf("add to the generic list of props >>> %s:\n",pname);)

  strncpy(pprop[0].name,"DESC",16);
  strncpy(pprop[0].desc,"Description",32);
  pprop[0].fmt = getTINEformat(DBF_STRING);
  pprop[0].elem = 40;

  strncpy(pprop[1].name,"VAL",16);
  strncpy(pprop[1].desc,"Value",32);
  pprop[1].fmt = getTINEformat(DBF_DOUBLE);
  pprop[1].elem = 1;

  strncpy(pprop[2].name,"EGU",16);
  strncpy(pprop[2].desc,"Engineering Units",32);
  pprop[2].fmt = getTINEformat(DBF_STRING);
  pprop[2].elem = 16;

  strncpy(pprop[3].name,"EGUF",16);
  strncpy(pprop[3].desc,"Engineering Units Full",32);
  pprop[3].fmt = getTINEformat(DBF_FLOAT);
  pprop[3].elem = 1;

  strncpy(pprop[4].name,"EGUL",16);
  strncpy(pprop[4].desc,"Engineering Units Low",32);
  pprop[4].fmt = getTINEformat(DBF_FLOAT);
  pprop[4].elem = 1;

  return(TINE_FREE_PROPERTIES);
}

long SplitRecordNames(NAME32 *prn, Tproperty *prp, long recNr)
{
  NAME32 plocName;
  char sp=1;
  char *c;
  int i,j=0;
  char desc[64],segu[8],dn[16],pn[32];
  float fmin,fmax;
  devChain *dc;
  int count, *pCount = &count;

  *pCount = 1;
  for(i = 0 ; i<recNr;i++) 
  { 
    strncpy(plocName.name, prn[i].name, 32);
    /* printf("plocName.name = %s\n", plocName.name); */
    for (c=strrev(plocName.name); c != NULL; c++)
    {
      if ((c=strchr(plocName.name,':')) != NULL && strlen(c) < 16) break;
      if ((c=strchr(plocName.name,'_')) != NULL && strlen(c) < 16) break;
    }

    /* printf("c = %d\n", c); */
    if (c != NULL)
    {
      sp = *c; *c++ = 0;
      strncpy(dn,strrev(c),16);
    }
    strncpy(pn,strrev(plocName.name),32);
    for (j=0; j<i && strncmp(pn,prp[j].name,32); j++);
    strncpy(prp[j].name,pn,32);   /* property name */
    prp[j].separater[0] = sp;        /* character separater (':', '_', or '\0') */
    if (sp != 0)
    {
      if ((dc=(devChain *)calloc(1,sizeof(devChain))) == NULL) return -1;
      strncpy(dc->name,dn,16);
      dc->next = prp[j].devs;
      prp[j].devs = dc;
      prp[j].ndev++;
    }
    
    tineGetFieldValue(prn[i].name,"DESC",desc,CF_TEXT,32,pCount);
    tineGetFieldValue(prn[i].name,"EGU", segu,CF_TEXT,5,pCount);
    tineGetFieldValue(prn[i].name,"EGUF",&fmax,CF_FLOAT,1,pCount);
    tineGetFieldValue(prn[i].name,"EGUL",&fmin,CF_FLOAT,1,pCount);
    sprintf(prp[j].desc,"[%g:%g %s]%s",fmin,fmax,segu,desc);
    
    /*
      some text info:
      strcpy(desc,"test_description");
      sprintf(prp[j].desc,"[0:0 S]%s",desc);
    */
     if(debug) printf("prp[%ld].desc='%s'\n",j, prp[j].desc);
  }
  /* j is the highest index found, so number of indices is j+1 ! */
  return(j+1);
}

int tineGetFieldValue(char *pname, char *pprop, void *ptr, int tineFmt, int len, int *pCount)
{
  long 		buffer[2000];
  long 		*pbuffer=&buffer[0];
  DBADDR 		addr;
  long		status = 0;
  long		options,nr;
  int		getUnixTime = FALSE;
  char		locName[40] = "";
  int		i,fmtsiz;

  DZK(printf(">> tineGetFieldValue(%s, %s, %x, %d, %d, %d)\n",pname,pprop,(int)ptr,tineFmt,len,(int)pCount);)
  sprintf(locName, "%s.%s", pname, pprop);

  if (ptr == NULL) {
    if(debug) printf("tineGetFieldValue for pname has bad NULL ptr , return");
    return -1;
  }
  fmtsiz = fmtsizeof(tineFmt);
  memset(ptr,0,fmtsiz*len);

  status=dbNameToAddr(&locName[0],&addr);
  if(status) {
    DZK(errMessage(status," dbNameToAddr failed");
	      printf(" dbNameToAddr failed >>> no record with name %s\n", locName);)
    return(ERRORnoChannel);
  }

  if( tineFmt == CF_LONG && addr.precord->dtyp == 2 ) getUnixTime = TRUE;
  if( getUnixTime ) {
	  sprintf(locName, "%s.%s", pname, "SVAL");
        
    DZK(printf("for Event Time String locName = %s\n",locName);)
    status=dbNameToAddr(&locName[0],&addr);
    if(status) {
      if(debug) errMessage(status," dbNameToAddr failed");
      if(debug) printf(" dbNameToAddr failed >>> no record with name %s\n", locName);
      return(ERRORnoChannel);
    }
  } 

  nr=MIN(addr.no_elements,((sizeof(buffer))/addr.field_size));
  options=0;
  if(addr.dbr_field_type==DBR_ENUM) {
    status=dbGetField(&addr,DBR_STRING,pbuffer,&options,&nr,NULL);
  } else {
    status=dbGetField(&addr,addr.dbr_field_type,pbuffer,&options,&nr,NULL);
  }
    
  if( status != 0 ) {
    if(debug) printf(" dbGetField failed with status = %d\n", (int)status);
    return(ERRORnoValue);
  }

  pCount = pCount + nr;
  DZK(printf("\t%s >>> addr.dbr_field_type = %d\n", locName, addr.dbr_field_type );)

  if(nr > 1) { 
    for(i = 0; i < nr; i++) {
      switch (addr.dbr_field_type) {
        case DBR_ENUM:
        case DBR_STRING:
        case DBR_CHAR:
        case DBR_UCHAR:
	        switch (tineFmt) {
	          case CF_SHORT: 	*(short *)ptr  = atoi((char *)buffer); ptr = (char *)ptr + sizeof(short); status = 0; break;
	          case CF_LONG: 	*(long *)ptr   = atoi((char *)buffer); ptr = (char *)ptr + sizeof(long); status = 0; break;
	          case CF_FLOAT: 	*(float *)ptr  = atof((char *)buffer); ptr = (char *)ptr + sizeof(float); status = 0; break;
	          case CF_DOUBLE: *(double *)ptr = atof((char *)buffer); ptr = (char *)ptr + sizeof(double); status = 0; break;
	          case CF_NAME16: strncpy(ptr, (char *)buffer, 16); ptr = (char *)ptr + 16; status = 0; break;
     		    case CF_NAME32: strncpy(ptr, (char *)buffer, 32); ptr = (char *)ptr + 32; status = 0; break;
            case CF_TEXT: 	strncpy(ptr, (char *)buffer, len); ptr = (char *)ptr + len; status = 0; break;
            default: 	status = illegal_format;
	        }
	        pbuffer += addr.field_size;	
	        break;
        case DBR_SHORT:
        case DBR_USHORT:
	        switch (tineFmt) {
	          case CF_SHORT:  *(short *)ptr  = *(short *)pbuffer; ptr = (char *)ptr + sizeof(short); status = 0; break; 
	          case CF_LONG:   *(long *)ptr   = *(short *)pbuffer; ptr = (char *)ptr + sizeof(long); status = 0; break;
	          case CF_FLOAT:  *(float *)ptr  = *(short *)pbuffer; ptr = (char *)ptr + sizeof(float); status = 0; break;  
	          case CF_DOUBLE: *(double *)ptr = *(short *)pbuffer; ptr = (char *)ptr + sizeof(double); status = 0; break;
	          case CF_NAME16: strncpy(ptr, (char *)buffer, 16); ptr = (char *)ptr + 16; status = 0; break;
	          case CF_NAME32: strncpy(ptr, (char *)buffer, 32); ptr = (char *)ptr + 32; status = 0; break;
	          case CF_TEXT:   strncpy(ptr, (char *)buffer, len); ptr = (char *)ptr + len; status = 0; break;
	          default:	    status = illegal_format;
	        }
	        pbuffer += addr.field_size;
	        break;
        case DBR_LONG:
        case DBR_ULONG:
	        switch (tineFmt) {
	          case CF_SHORT:  *(short *)ptr  = *pbuffer; ptr = (char *)ptr + sizeof(short); status = 0; break;
	          case CF_LONG:   *(long *)ptr   = *pbuffer; ptr = (char *)ptr + sizeof(long); status = 0; break;
	          case CF_FLOAT:  *(float *)ptr  = *pbuffer; ptr = (char *)ptr + sizeof(float); status = 0; break;
	          case CF_DOUBLE: *(double *)ptr = *pbuffer; ptr = (char *)ptr + sizeof(double); status = 0; break;
	          case CF_NAME16: strncpy(ptr, (char *)buffer, 16); ptr = (char *)ptr + 16; status = 0; break;
	          case CF_NAME32: strncpy(ptr, (char *)buffer, 32); ptr = (char *)ptr + 32; status = 0; break;
	          case CF_TEXT:   strncpy(ptr, (char *)buffer, len); ptr = (char *)ptr + len; status = 0; break;
	          default:	    status = illegal_format;
	        }
	        pbuffer += addr.field_size;
	        break;
        case DBR_FLOAT:
	        switch (tineFmt) {
	          case CF_SHORT:  *(short *)ptr  = *(float *)pbuffer; ptr = (char *)ptr + sizeof(short); status = 0; break;
	          case CF_LONG:   *(long *)ptr   = *(float *)pbuffer; ptr = (char *)ptr + sizeof(long); status = 0; break;
	          case CF_FLOAT:  *(float *)ptr  = *(float *)pbuffer; ptr = (char *)ptr + sizeof(float); status = 0; break;
	          case CF_DOUBLE: *(double *)ptr = *(float *)pbuffer; ptr = (char *)ptr + sizeof(double); status = 0; break;
	          case CF_NAME16: cvtFloatToString(*(float *)pbuffer, ptr, NULL); ptr = (char *)ptr + 16; status = 0; break;
	          case CF_NAME32: cvtFloatToString(*(float *)pbuffer, ptr, NULL); ptr = (char *)ptr + 32; status = 0; break;
	          case CF_TEXT:   cvtFloatToString(*(float *)pbuffer, ptr, NULL); ptr = (char *)ptr + len; status = 0; break;
	          default:	    status = illegal_format;
	        }
	        pbuffer += addr.field_size;
	        break;
        case DBR_DOUBLE:
	        switch (tineFmt) {
            case CF_SHORT:  *(short *)ptr  = *(double *)pbuffer; ptr = (char *)ptr + sizeof(short); status = 0; break;
            case CF_LONG:   *(long *)ptr   = *(double *)pbuffer; ptr = (char *)ptr + sizeof(long); status = 0; break;
            case CF_FLOAT:  *(float *)ptr  = *(double *)pbuffer; ptr = (char *)ptr + sizeof(float); status = 0; break;
            case CF_DOUBLE: *(double *)ptr = *(double *)pbuffer; ptr = (char *)ptr + sizeof(double); status = 0; break;
            case CF_NAME16: cvtDoubleToString(*(double *)pbuffer, ptr, NULL); ptr = (char *)ptr + 16; status = 0; break;
            case CF_NAME32: cvtDoubleToString(*(double *)pbuffer, ptr, NULL); ptr = (char *)ptr + 32; status = 0; break;
            case CF_TEXT:   cvtDoubleToString(*(double *)pbuffer, ptr, NULL); ptr = (char *)ptr + len; status = 0; break;
            default:        status = illegal_format;
          }
	        pbuffer += addr.field_size;
	        break;
        case DBR_PUT_ACKT:
        case DBR_PUT_ACKS:
        case DBR_NOACCESS:
        default: 		
	        DZK(printf("default\n");)
	        status = ERRORnoValue;
      }
    }
  } else {
    switch (addr.dbr_field_type) {
        case DBR_ENUM:
        case DBR_STRING:
        case DBR_CHAR:
        case DBR_UCHAR:
          DZK(printf("DBR_STRING %s \n", (char *)buffer);)
          switch (tineFmt) {
            case CF_SHORT:  *(short *)ptr  = atoi((char *)buffer); status = 0; break;
            case CF_LONG:   *(long *)ptr   = atoi((char *)buffer); status = 0; break;
            case CF_FLOAT:  *(float *)ptr  = atof((char *)buffer); status = 0; break; 
            case CF_DOUBLE: *(double *)ptr = atof((char *)buffer); status = 0; break;
            case CF_NAME16: strncpy(ptr, (char *)buffer, 16); status = 0; break;
            case CF_NAME32: strncpy(ptr, (char *)buffer, 32); status = 0; break;
            case CF_TEXT:   strncpy(ptr, (char *)buffer, len); status = 0; break;
            default: 	    status = illegal_format;
          }
          break;
        case DBR_SHORT:
        case DBR_USHORT:
	        DZK(printf("DBR_SHORT %d\n", *(short *)pbuffer);)
	        switch (tineFmt) {
	          case CF_SHORT:  *(short *)ptr  = *(short *)pbuffer; status = 0; break; 
	          case CF_LONG:   *(long *)ptr   = *(short *)pbuffer; status = 0; break;
	          case CF_FLOAT:  *(float *)ptr  = *(short *)pbuffer; status = 0; break;  
	          case CF_DOUBLE: *(double *)ptr = *(short *)pbuffer; status = 0; break;
	          case CF_NAME16: strncpy(ptr, (char *)buffer, 16); status = 0; break;
	          case CF_NAME32: strncpy(ptr, (char *)buffer, 32); status = 0; break;
	          case CF_TEXT:   strncpy(ptr, (char *)buffer, len); status = 0; break;
	          default:	    status = illegal_format;
	        }
	        break;
        case DBR_LONG:
        case DBR_ULONG:
	        DZK(printf("DBR_LONG %d\n", *(int *)pbuffer);)
	        switch (tineFmt) {
	          case CF_SHORT:  *(short *)ptr = *pbuffer; status = 0; break;
	          case CF_LONG:   *(long *)ptr = *pbuffer; status = 0; break;
	          case CF_FLOAT:  *(float *)ptr = *pbuffer; status = 0; break;
	          case CF_DOUBLE: *(double *)ptr = *pbuffer; status = 0; break;
	          case CF_NAME16: strncpy(ptr, (char *)buffer, 16); status = 0; break;
	          case CF_NAME32: strncpy(ptr, (char *)buffer, 32); status = 0; break;
	          case CF_TEXT:   strncpy(ptr, (char *)buffer, len); status = 0; break;
	          default:	    status = illegal_format;
	        }
	        break;
        case DBR_FLOAT:
	        DZK(printf("DBR_FLOAT %f\n", *(float *)pbuffer);)
	        switch (tineFmt) {
	          case CF_SHORT:  *(short *)ptr = *(float *)pbuffer; status = 0; break;
	          case CF_LONG:   *(long *)ptr = *(float *)pbuffer; status = 0; break;
	          case CF_FLOAT:  *(float *)ptr = *(float *)pbuffer; status = 0; break;
	          case CF_DOUBLE: *(double *)ptr = *(float *)pbuffer; status = 0; break;
	          case CF_NAME16:	
	          case CF_NAME32:
	          case CF_TEXT:   cvtFloatToString(*(float *)pbuffer, ptr, NULL); status = 0; break;
	          default:	    status = illegal_format;
	        }
	        break;
        case DBR_DOUBLE:
	        DZK(printf("DBR_DOUBLE %f\n", *(double *)pbuffer);)
	        switch (tineFmt) {
            case CF_SHORT:  *(short *)ptr = *(double *)pbuffer; status = 0; break;
            case CF_LONG:   *(long *)ptr = *(double *)pbuffer; status = 0; break;
            case CF_FLOAT:  *(float *)ptr = *(double *)pbuffer; status = 0; break;
            case CF_DOUBLE: *(double *)ptr = *(double *)pbuffer; status = 0; break;
            case CF_NAME16:
            case CF_NAME32:
            case CF_TEXT:   cvtDoubleToString(*(double *)pbuffer, ptr, NULL); status = 0; break;
            default:        status = illegal_format;
          }
          break;
        case DBR_PUT_ACKT:
        case DBR_PUT_ACKS:
        case DBR_NOACCESS:
        default: 		
	        DZK(printf("default\n");)
	        status = ERRORnoValue;
    }
  }

  return(status);
}

long tinePutFieldValue(char *pname,char *pvalue)
{
  /* declare buffer long just to ensure correct alignment */
  DBADDR addr;
  long	  status;

  /* make sure value was passed*/
  if(pvalue==0) {
    if(debug) errMessage(0L,"No value was specified");
    return(1);
  }
  status=dbNameToAddr(pname,&addr);
  if(status) {
    if(debug) errMessage(status,"dbNameToAddr error");
    return(status);
  }
  /* For enumerated types must allow for ENUM rather than string*/
  /* If entire field is digits then use DBR_ENUM else DBR_STRING*/
  if((addr.dbr_field_type==DBR_ENUM) && (*pvalue!=0)
      &&  (strspn(pvalue,"0123456789")==strlen(pvalue))) {
	  short value;

	  sscanf(pvalue,"%hu",&value);
	  status=dbPutField(&addr,DBR_ENUM,&value,1L);
  } else {
    status=dbPutField(&addr,DBR_STRING,pvalue,1L);
  }
  if(status) {
    if(debug) errMessage(status,"dbPutField error");
    return(status);
  }
  return(status);
}
#ifndef EPICS_3_13
int tineGetAlarmLimits(char *recName, E2T_ALARM_XREF *pXRef)
{
  int *pGetVal = (int*)1;
  int sv,status;
  char lsv[16],llsv[16],hsv[16],hhsv[16];
  char *plsv = &lsv[0];
  char *pllsv = &llsv[0];
  char *phsv = &hsv[0];
  char *phhsv = &hhsv[0];

  if (recName == NULL || pXRef == NULL) 
  {
    if(debug) printf("tineGetAlarmLimits: input error -> NULL pointer as argument!!!\n");
    return FALSE;
  }
  if(tineGetFieldValue(recName,"LLSV",plsv,CF_TEXT,1,pGetVal) &&
     tineGetFieldValue(recName,"LSV",pllsv,CF_TEXT,1,pGetVal) &&
     tineGetFieldValue(recName,"HSV",phsv,CF_TEXT,1,pGetVal) &&
     tineGetFieldValue(recName,"HHSV",phhsv,CF_TEXT,1,pGetVal))
    { /* this channel does not have an alarm record */
       if(debug) printf("tineGetAlarmLimits:does not have an alarm record!!!\n");
      return FALSE; 
    }

  /* try to map EPICS severity into TINE severity :*/
  pXRef[0].severity = 12; pXRef[0].code = 0;
  strcpy(pXRef[0].text,"record value is near threshold");
  strcpy(pXRef[0].tag,"warn too low");
  pXRef[1].severity = 14; pXRef[1].code = 1;
  strcpy(pXRef[1].text,"record value has exceeded threshold !");
  strcpy(pXRef[1].tag,"value too low");
  pXRef[2].severity = 12; pXRef[2].code = 2;
  strcpy(pXRef[2].text,"record value is near threshold");
  strcpy(pXRef[2].tag,"warn too high");
  pXRef[3].severity = 14; pXRef[3].code = 3;
  strcpy(pXRef[3].text,"record value has exceeded threshold !");
  strcpy(pXRef[3].tag,"value too high");
  /* how to get the following information ? */
  pXRef[0].system = 1200;
  pXRef[1].system = 1200;
  pXRef[2].system = 1200;
  pXRef[3].system = 1200;
  strcpy(pXRef[0].url,"http://www-kryo.desy.de");
  strcpy(pXRef[1].url,"http://www-kryo.desy.de");
  strcpy(pXRef[2].url,"http://www-kryo.desy.de");
  strcpy(pXRef[3].url,"http://www-kryo.desy.de");
  status = e2t_add_event(recName, &sv);
  if(debug) if(status) printf("e2t_add_event: recName = %s\n",recName);
  return TRUE;
}

int tineSetAlarm(char *chname,int sev,int status,float fval)
{
  char devName[32];
  long almCode=0;
  int devnr;
  ADS *ads=NULL;
  DZK(printf("tineSetAlarm Epicsname =%s\n",chname););

  if (makeDeviceNameFromChannelName(chname,devName) != 0) strcpy(devName,"#0");
  if ((devnr=GetDeviceNumber("GTWEQM",devName)) < 0) devnr = 0;
  switch (status)
  {
    case LOW_ALARM: almCode = 0; break;
    case LOLO_ALARM: almCode = 1; break;
    case HIGH_ALARM: almCode = 2; break;
    case HIHI_ALARM: almCode = 3; break;
  default: {
      DZK(printf("tineSetAlarm Unpredictable status= %d\n",status);)
      return 0; 
    }
  }
  almCode += USR_ALARM_OFFSET + devnr*4;

  if ((ads=getAlarmTable("GTWEQM",almCode)) != NULL)
  {
    ads->severity = sev == MAJOR_ALARM ? 14 : 12;

  }
  if (sev > 0){ 
    DZK(printf("tineSetAlarm  positive sev=%d\n",sev););
    SetAlarm("GTWEQM",devnr,almCode,(BYTE *)&fval); 
  } else { 
    DZK(printf("tineSetAlarm  ZERO sev=%d\n",sev););
    RemoveAlarm("GTWEQM",devnr,almCode);
  }

  return(0);
}
#endif
int tineGetRecordFromFile(void)
{
  int c;
  FILE *fp = NULL;
  char str[32];
  static char RecordFilename[32];

  if (strlen(RecordFilename) == 0) {
    sprintf(RecordFilename,"%s","freeRecord.list");
    if(debug) printf("file to open = %s\n", RecordFilename);
 	}
  if ((fp=fopen(RecordFilename,"r")) == NULL) {
    if(debug) printf("no such file\n");
    return -1;
 	}

  c = 0;
  while (fscanf(fp,"%s",str) != EOF) {
    if (strchr("#",str[0])) continue;
    c++;
 	}
  if (feof(fp)) fclose(fp);
  return c;
}

int getPropsNotInclude(void)
{
  int c;
  FILE *fp = NULL;
  char str[8];
  static char PropsFilename[32];

  if (strlen(PropsFilename) == 0) {
    sprintf(PropsFilename,"%s","propsNotIncl.list");
    if(debug) printf("file to open = %s\n", PropsFilename);
 	}
  if ((fp=fopen(PropsFilename,"r")) == NULL) {
    if(debug)  printf("no such file\n");
    return -1;
 	}

  c = 0;
  while (fgets(str,7,fp) != NULL) {
    if (strchr("#",str[0])) continue;      /* comment line(s) */
    if(debug) printf("%d\t%s\n",c,str);
    c++;
 	}

  if (feof(fp)) fclose(fp);
  return 1;
}

char *makeCaName(char *devName,int prpId)
{    
  static char name[64];
  int i;
  char *c;

  name[0] = 0;
  if (prpId < 2*numProperties)
  {
    i = prpId % numProperties;
    if (strlen(devName) > 0 && devName[0] != '#')
    {
      strncpy(name,devName,16); name[16] = 0;
      strcat(name,pT[i].separater);
    }
    strncat(name,pT[i].name,32);
    /* strncat(name,pT[i].postfix,8);  */ /* <- is this necessary ? */
  }
  else /* "VAL" or "DESC" */
  {
    strncpy(name,devName,32);
    if ((c=strchr(&name[1],'#')) != NULL)
    { /* caller is sending a 'registered' TINE device instead of a channel access device */
      *c++ = 0;
      i = atoi(c);
      strcat(name,pT[i].separater);
      strncat(name,pT[i].name,32);
    }
  }
  return name;
}
char *makeCaField(int prpId)
{
  if (prpId < numProperties) return "VAL";
  if (prpId < 2*numProperties) return "NAM";
  if (prpId == 2*numProperties) return "VAL";
  if (prpId == 2*numProperties+1) return "DESC";
  if (prpId == 2*numProperties+2) return "RVAL";
  return "VAL";
}
/****************************************************************
 *								*
 *	TINE Specific Part					*
 *								*
 ****************************************************************/
int readStructureFile(char *filename, NAME32 **cmpPrps, NAME32 **cmpDevPrps, NAME32 **cmpDevs, NAME16 **cmpAliasDevs, NAME32 **cmpDesc, int *numCmpProperties)
{
int cc=0,i = 0,n = 0,siz = 0,done = 0,count = 0;
char str[256],hdr[256];
int prp_col=0,dev_col=0,alias_col=0, desc_col=0;
FILE *fp = NULL;
NAME32 *devprps = NULL;
NAME32 *prps = NULL;
NAME32 *devs = NULL;
NAME16 *aliasdevs = NULL;
NAME32 *desc = NULL;
int openCount;

	do {
		siz=GetDataBaseSize(filename);
		count++;
		if(debug)printf("reading composite structure file <%d> %s\n",count,filename);
	} while (siz <= 0 && count <= 5 );
	if (siz <=0 && count == 6) ccerr(no_such_file);

	*numCmpProperties = 0;
	openCount = 0;
	while(fp == NULL && openCount <= 3) {
		if ((fp=fopen(filename,"r")) != NULL) {
			if ((devprps=(NAME32 *)calloc(siz,sizeof(NAME32))) == NULL) ccerr(out_of_local_memory);
			if ((prps=(NAME32 *)calloc(siz,sizeof(NAME32))) == NULL) ccerr(out_of_local_memory);
			if ((devs=(NAME32 *)calloc(siz,sizeof(NAME32))) == NULL) ccerr(out_of_local_memory);
			if ((aliasdevs=(NAME16 *)calloc(siz,sizeof(NAME16))) == NULL) ccerr(out_of_local_memory);
			if ((desc=(NAME32 *)calloc(siz,sizeof(NAME32))) == NULL) ccerr(out_of_local_memory);
			for (done=0; fgets(str,255,fp) != NULL;) {
			if (strlen(str) == 0) continue;          /* carriage return */
			if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
			if (!done) {
				strcpy(hdr,str);
				if ((dev_col=findcol(hdr,"DEVICE")) < 0) ccerr(no_such_column);
				if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_column);
				if ((alias_col=findcol(hdr,"ALIAS")) < 0) ccerr(no_such_column);
				desc_col=findcol(hdr,"DESCRIPTION");
				done = 1;
				continue;
			}
			strncpy(devprps[i].name,colptr(prp_col,str),32);
			strncpy(devs[i].name,colptr(dev_col,str),32);
			strncpy(aliasdevs[i].name,colptr(alias_col,str),16);
			if(desc_col >= 0) strncpy(desc[i].name,colptr(desc_col,str),32);
			else strcpy(desc[i].name,"");
			for (n=0; n<i && strncmp(devprps[n].name,devprps[i].name,32); n++);
				if (n == i) {
					strncpy(prps[*numCmpProperties].name,devprps[i].name,32);
					*numCmpProperties = *numCmpProperties + 1;
				}
				i++;
			}
			*cmpDevPrps = devprps; *cmpDevs = devs;
			*cmpPrps = prps; *cmpAliasDevs = aliasdevs;
			*cmpDesc = desc;
			fclose(fp);
			return i;
		}
		openCount++;
	}
	err:
	if(debug) printf("error read struct file\n");
	if(debug) if (cc) printf("read structure file : %s\n",erlst[cc]);
	return 0;
}

int readCompositeFile(char *cmpFileName)
{
  int i,cc = 0,done = 0;
  char str[256],hdr[256],scratch[32];
  int name_col=0,fmt_col=0,file_col=0,desc_col=0,devs_col=0;
  FILE *fp = NULL;
  int compCount=0;

  DZK(printf("reading composite file %s\n",cmpFileName);)
  while(ncomposites <= 0 && compCount <= 3) 
  {
    if ((ncomposites=GetDataBaseSize(cmpFileName)) <= 0) ccerr(no_such_file);
    compCount++;
 	}

  compCount=0;
  while(fp == NULL && compCount <= 3) 
  {
    if ((fp=fopen(cmpFileName,"r")) != NULL) 
    {
      if ((composites=(COMPOSITE *)calloc(ncomposites,sizeof(COMPOSITE))) == NULL) ccerr(out_of_local_memory);
      for (done=0,i=0; fgets(str,255,fp) != NULL;) 
      {
        if (strlen(str) == 0) continue;          /* carriage return */
        if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
        if (!done) 
        {
          strcpy(hdr,str);
          if ((name_col=findcol(hdr,"PROPERTYNAME")) < 0) ccerr(no_such_column);
          if ((fmt_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_column);
          if ((file_col=findcol(hdr,"STRUCTUREFILE")) < 0) ccerr(no_such_column);
          devs_col = findcol(hdr,"DEVICENAME");
          desc_col=findcol(hdr,"DESCRIPTION");
          done = 1;
          continue;
        }
        if (devs_col >= 0) strncpy(composites[i].devName,colptr(devs_col,str),32);
        strncpy(composites[i].devProperty,colptr(name_col,str),32);
        strncpy(scratch,colptr(fmt_col,str),32);
        composites[i].prpFormat = ftoi(scratch);
        strncpy(scratch,colptr(file_col,str),32);
        composites[i].prpLength = readStructureFile(scratch, &composites[i].cmpPrps, &composites[i].cmpDevPrps, &composites[i].cmpDevs, &composites[i].cmpAliasDevs, &composites[i].cmpDesc, &composites[i].numCmpPrps);
        if(desc_col >= 0) strncpy(composites[i].prpDesc,colptr(desc_col,str),32);
        else strcpy(composites[i].prpDesc,"");
        i++;
      }
      if (i < ncomposites) ncomposites = i;
      if(debug >= 1) 
      {
        printf("Composites\t\tdevProperty\n");
        for(i=0;i<ncomposites;i++) 
        {
          printf("%s\t\t", (char *)composites[i].cmpDevs);
          printf("%s\n", composites[i].devProperty);
        }
      }
      fclose(fp);
      return 0;
		}
    compCount++;
  }
  cc = no_such_file;
err:
  if(debug)  printf("ERROR >>>Composite file : %s\n",erlst[cc]);
  return cc;
}

short getTINEformat(int epicsFmt)
{
	switch (epicsFmt) {
  	case DBF_STRING:	return CF_TEXT;
  	case DBF_CHAR:
  	case DBF_UCHAR:		return CF_BYTE; 
  	case DBF_SHORT:	
  	case DBF_USHORT:	return CF_SHORT;
  	case DBF_LONG:
  	case DBF_ULONG:		return CF_LONG;
  	case DBF_FLOAT:		return CF_FLOAT;
  	case DBF_DOUBLE:	return CF_DOUBLE;
  	case DBF_ENUM:		return CF_LONG;
  	case DBF_MENU:		return CF_NAME16;
  	case DBF_DEVICE:	return CF_TEXT;
  	case DBF_INLINK:	return CF_NAME32;
  	case DBF_OUTLINK:	return CF_NAME32;
  	case DBF_FWDLINK:	return CF_NAME32;
  	case DBF_NOACCESS:	return CF_NULL;
  	default: 		return CF_DOUBLE;
  }
}

int fillDeviceNamesAndProperties(char *fec,char *server,NAME32 **devnames,int *numDevices,Tproperty **prp,int *numProps)
{
  int cc=0,i;
  NAME32 *pD;

  /* fill in fec and server */
  strcpy(server, fec);

  *numDevices = tineGetNumRecords();
  if(debug) printf("numDevices : %d\n",*numDevices);

  if(*numDevices <= 0) return no_such_file;
  if ((pD=(NAME32 *)calloc(*numDevices,sizeof(NAME32))) == NULL) ccerr(out_of_server_memory);
  if (tineGetRecords(pD) != *numDevices) ccerr(database_not_loaded);

  DZK(for (i=0; i<*numDevices; i++) {
    if(i<10) printf("%s\t", pD[i].name);
  }
  printf("\nTotal of Records: %d\n", *numDevices);)

  if(*numDevices > 0) {
    if((pT=(Tproperty *)calloc(*numDevices,sizeof(Tproperty))) == NULL) ccerr(out_of_local_memory);
    DZK(if(debug) printf("pT = %x\n",(int)pT);)
  }

  /* Split the Record Names */
  *numProps = SplitRecordNames(pD, pT, *numDevices);
  if(debug) printf("*numProps = %d\n", *numProps);

  *prp = pT;
  *devnames = pD;
  if(debug) printf("properties of %s:\n",server);

err:
  return(cc);
}

int gtweqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int nr,len=0,cc=0;
  int prpId;
  char caTempName[32];	
  char caName[32];	
  char caField[8];
  char errorCA[32];	
  int i,status;
  int cmpId = -1;
  char epicsProperty[32];
  char *c,tmpPrpName[32];
  void *ptr;
  int *pCount = 0;
  UINT32 bufsiz;
  char strbuf[40];
  char *pStrBuf = &strbuf[0];
  int nrSetComp = 3;	/* Number of composite sets Val, Val.NAME, Val.DESC */
  devChain *dc;
  int isMetaProperty = FALSE;

  DZK(printf("gtweqm params: devName = %s\tdevProperty = %s\n",devName,devProperty);)

  if ((prpId=GetPropertyId("GTWEQM",devProperty)) < 0) return illegal_property;
  if (prpId >= offsetComposites) 
  { /* asking for a composite ... */
    if (prpId > offsetComposites + ncomposites*nrSetComp) ccerr(illegal_property);
    cmpId = prpId - offsetComposites;
    if (access & CA_WRITE)
    { /* fill this in in case of composite WRITE commands ... */
      nr = GetDeviceNumberEx("GTWEQM",devName,devProperty); 
      if (nr < 0) return illegal_equipment_number;
      strncpy(caName,composites[cmpId].cmpDevs[nr].name,32);
      strncpy(caField,"VAL",8);
    }
  } 
  else 
  {
    strncpy(caName,makeCaName(devName,prpId),32);
    strncpy(caField,makeCaField(prpId),8);
  }
  DZK(printf("prpId = %d\tcmpId = %d\n",prpId,cmpId);)

  if (access & CA_WRITE) if (din->dArrayLength == 0) return 0;

  /* zu verbessern */
  strncpy(tmpPrpName,devProperty,16);
  /* change .NAME to .NAM for systematics ... */
  if ((c=strstr(tmpPrpName,".NAM")) != NULL) { *c = 0; isMetaProperty = TRUE; }
  if ((c=strstr(tmpPrpName,".DESC")) != NULL) { *c = 0; isMetaProperty = TRUE; }

  if (cmpId > -1) 
  {
    for(i=0;i<ncomposites;i++) 
    {
      if (strncmp(composites[i].devProperty,tmpPrpName,32)) continue;
      break;
    }
    if (i == ncomposites) return illegal_property;
  }

  if (access & CA_WRITE) 
  {
    strncpy(caTempName, caName, 32);
    strcat(caTempName,".");
    strcat(caTempName,caField);
    DZK(printf("name = %s\tval = %s\n", caTempName, din->data.cptr);)
    bufsiz = sizeof(strbuf);
    memset(strbuf,0,bufsiz);
    cc = getValuesAsString(din,pStrBuf,&bufsiz);
    if(cc) return(cc);
    DZK(printf("return strbuf = %s\n", strbuf);)
    if ((cc=tinePutFieldValue(caTempName, pStrBuf)) != 0)
    {
      SetRPCCompletion("Channel Error");
      return 512;
    }
  }
  DZK(printf("cmpId = %d\tncomposites = %d\n",cmpId,ncomposites);
      printf("dout->dArrayLength = %d\n",(int)dout->dArrayLength);)
  if ((len=dout->dArrayLength) > 0) 
  { 
    if (cmpId >= 0 && cmpId < ncomposites*nrSetComp) 
    { /* composite property ... */
      if (dout->dFormat == CF_TEXT) return illegal_format;
      DZK(printf("got cmpId [%d]\n", cmpId);)
      /* ADDED P.D. */
      if (strstr(devProperty,".NAM") != NULL) 
      {
        if (dout->dFormat != CF_NAME16) return illegal_format;
        cmpId -= ncomposites;
        for (i=0; i<composites[cmpId].prpLength && i<dout->dArrayLength; i++) 
        {
          memcpy(((NAME16 *)dout->data.vptr)[i].name, composites[cmpId].cmpAliasDevs[i].name, 32);
        }
        return 0;
      }
      if (strstr(devProperty,".DESC") != NULL) 
      {
        if (dout->dFormat != CF_NAME32) return illegal_format;
	      cmpId -= ncomposites*2;
        for (i=0; i<composites[cmpId].prpLength && i<dout->dArrayLength; i++) 
        {
          memcpy(((NAME32 *)dout->data.vptr)[i].name, composites[cmpId].cmpDesc[i].name, 32);
        }
        return 0;
      }

      if (cmpId >= ncomposites) return illegal_property; /* can no longer be true ... */
      DZK(printf("dout->dArrayLength: %d\n", (int)dout->dArrayLength);)
      for (i=0; i<composites[cmpId].prpLength && i<dout->dArrayLength; i++) 
      {
        strncpy(epicsProperty,composites[cmpId].cmpDevPrps[i].name,32);
        strncpy(caName,composites[cmpId].cmpDevs[i].name,32);
        /* request for channel names */
        if ((dout->dFormat == CF_NAME32) && (strcmp(composites[cmpId].cmpDevPrps[i].name, "NAME") == 0)) 
        {
          memcpy(((NAME32 *)dout->data.vptr)[i].name, caName, 32);
          continue;	
        }

        ptr = dout->data.bptr + i*fmtsizeof(dout->dFormat);

       status = tineGetFieldValue(&caName[0], epicsProperty, ptr, dout->dFormat, dout->dArrayLength, pCount);
        if(status) 
        {
          sprintf(errorCA,"CA Name %s\n", caName);
          SetRPCCompletion(errorCA);
          return 512;
        }
      }
    } 
    else 
    { /* not a composite ... */
      if (strstr(devProperty,".NAM") != NULL)
      {
        nr = prpId % numProperties;
        for (i=0,dc=pT[nr].devs; i<dout->dArrayLength && dc != NULL; i++, dc=dc->next)
          strncpy(((NAME16 *)dout->data.vptr)[i].name,dc->name,16);
        dout->dArrayLength = i;
      }
      else
      {
        status = tineGetFieldValue(&caName[0], caField, dout->data.vptr, dout->dFormat, dout->dArrayLength, pCount);
        if(status) 
        {
          SetRPCCompletion("Channel Error");
          return 512;
        }
      }
    }
  }

err:
  if (cc) if(debug) printf("read structure file : %s\n",erlst[cc]);

  pCount = 0;
  return 0;
}

int E2T()
{
  int status,i,j,k,cc;
  int numDevices;
  char tmp[64], tmpDevName[64], devServer[64];
  NAME32 *devnames = NULL;
  Tproperty *prp = NULL;
  DTYPE dout, din;
# ifndef EPICS_3_13
  E2T_ALARM_XREF almXRef[4];
  ADS ads;
  devChain *dc=NULL;
  int severity = 0;
# endif
  
  AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);

  NGdebug = 0;
  /* find out if there are any composite properties ... */
  readCompositeFile("composites.csv");

  max_rcvbuf = 0;
  max_sndbuf = 0;
  if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }

  /* find out how many and the names of the devices and properties ... */
  status = fillDeviceNamesAndProperties(gFecName,devServer,&devnames,&numDevices,&prp,&numProperties);
  if(status) 
  {
    if(debug) printf("fillDeviceNamesAndProperties : <%d> %s\n",status,erlst[status]);
    return(-1);
  }

  /* register server name dynamically ... */
  /* RegisterFecName(fecName,"EPICS To TINE Translator","VXW","Bldg 52","??","P.Duval Z.Kakucs",0); */

  /* initialize the TINE server ... */
  /* longDeviceNames = TRUE;  <- shouldn't need this any more ! */
  /* register the equipment module ... */
  if(ncomposites <= 0) 
  {
     if(debug)  printf("No composites\n");
     ncomposites = 0;
  }
  if(debug) printf("%s %s %d devices & %d composites\n",gFecName,devServer,numDevices,ncomposites);
  RegisterEquipmentModule(devServer,"GTWEQM",numDevices+ncomposites,gtweqm,NULL,NULL,0,NULL);

  /* register all known properties ... */
  for (i=0; i<numProperties; i++) 
  {
    RegisterPropertyEx("GTWEQM",prp[i].name,prp[i].ndev,CF_DOUBLE,CA_READ,prp[i].desc,i);
  }
  for (i=0; i<numProperties; i++)
  {
    strncpy(tmp,prp[i].name,32);
    strcat(tmp,".NAM");
    RegisterPropertyEx("GTWEQM",tmp,prp[i].ndev,CF_NAME16,CA_READ,prp[i].desc,i+numProperties);
  }

  /* be backwards compatible ... */
  RegisterPropertyEx("GTWEQM","VAL",1,CF_DOUBLE,CA_READ,"epics value",2*numProperties);
  RegisterPropertyEx("GTWEQM","DESC",40,CF_TEXT,CA_READ,"epics description",2*numProperties+1);
  RegisterPropertyEx("GTWEQM","RVAL",1,CF_DOUBLE,CA_READ,"epics r-value",2*numProperties+2);

  offsetComposites = 2*numProperties + 3;

  DZK(printf("composites summe: %d\n", ncomposites);)
  for (i=0; i<ncomposites; i++) 
  {
    dout.dArrayLength = composites[i].prpLength;
    dout.dFormat = composites[i].prpFormat;
    din.dArrayLength = 0; din.dFormat = CF_NULL;
    RegisterPropertyInformation("GTWEQM",composites[i].devProperty,&dout,&din,CA_READ,AT_CHANNEL,0,composites[i].prpDesc,offsetComposites+i,"");
    /* RegisterPropertyEx("GTWEQM",composites[i].devProperty,composites[i].prpLength, composites[i].prpFormat,CA_READ,composites[i].prpDesc,offsetComposites+i); */
    sprintf(tmp,"%s.NAM",composites[i].devProperty);
    RegisterPropertyEx("GTWEQM",tmp,composites[i].prpLength, CF_NAME16, CA_READ,composites[i].prpDesc,offsetComposites+ncomposites+i);
    sprintf(tmp,"%s.DESC",composites[i].devProperty);
    RegisterPropertyEx("GTWEQM",tmp,composites[i].prpLength, CF_NAME32, CA_READ,composites[i].prpDesc,offsetComposites+2*ncomposites+i);
  }

# ifndef EPICS_3_13
  /* check for alarms ... */
  /* was i<=numProperties !!!! */
  for (i=0,j=0; i<numProperties; i++) for (dc=prp[i].devs; dc != NULL; dc=dc->next)
  {
    strncpy(tmp,dc->name,16); tmp[16] = 0;
    strcat(tmp,prp[i].separater);
    strncat(tmp,prp[i].name,32);
    DZK(printf("before tineGetAlarmLimits: recName = %s\n", tmp);)
    if(tineGetAlarmLimits(tmp, almXRef)) 
    {
      for (k=0; k<4; k++)
      {
        ads.alarmCode = USR_ALARM_OFFSET + j*4 + almXRef[k].code;
        ads.alarmMask = 0xff;
        ads.severity = almXRef[k].severity;
        ads.alarmDataFormat = CF_FLOAT;
        ads.alarmDataArraySize = 1;
        ads.alarmSystem = almXRef[k].system; 
        strncpy(ads.alarmTag,almXRef[k].tag,16);
        strncpy(ads.alarmText,almXRef[k].text,40);
        strncpy(ads.dataText,"data value which triggered the alarm",40);
        strncpy(ads.url,almXRef[k].url,40);
        strncpy(ads.deviceText,tmp,38);
        appendAlarmInfoTable("GTWEQM",&ads);
      }
    }
    makeDeviceNameFromChannelName(tmp,tmpDevName);
    RegisterDeviceName("GTWEQM",tmpDevName,j++);  
  }
# else
  for (i=0; i<numDevices; i++) 
  {
    RegisterDeviceName("GTWEQM",devnames[i].name,i);
  }
# endif
  for (i=0; i<ncomposites; i++) 
  {
    DZK(printf("composites[%d].devName = %s\n", i, composites[i].devName);)
    RegisterDeviceName("GTWEQM",composites[i].devName,i+numDevices);
  }

  /* run the server ...   */
  for(;;) {
    SystemCycle(TRUE);
  }	
  return 0;
}
int makeDeviceNameFromChannelName(char *chName,char *devName)
{
  char *c,lclName[36],tmpName[32],pn[32];
  int pid = 0;
  if (chName == NULL || devName == NULL) return argument_list_error;
  strncpy(lclName,chName,32); lclName[32] = 0;
  if ((c=strstr(lclName,".VAL")) != NULL) *c = 0;
  if (strlen(lclName) <= 16) 
  {
    strncpy(devName,lclName,16); 
    devName[16] = 0;
  }
  else
  {
    strncpy(tmpName,lclName,32);
    for (c=strrev(tmpName); c != NULL; c++)
    {
      if ((c=strchr(tmpName,':')) != NULL && strlen(c) < 16) break;
      if ((c=strchr(tmpName,'_')) != NULL && strlen(c) < 16) break;
    }
    if (c != NULL)
    {
      *c++ = 0;
      strncpy(devName,strrev(c),12); 
      devName[12] = 0;
    }
    strncpy(pn,strrev(tmpName),32);
    if ((pid=GetPropertyId("GTWEQM",pn)) < 0) pid = 0;
    sprintf(pn,"#%d",pid);
    strncat(devName,pn,4);
  }
  return 0;
}

#ifdef vxWorks
void loadE2TModules()
{
  int status;
  MODULE_ID modId;
  char pathName[MAX_FILENAME_LENGTH];
  char *pPathName = &pathName[0];
  char pathFrom[MAX_FILENAME_LENGTH];
  /* char *pPathFrom = &pathFrom[0]; */
  char archName[8];
  char *pArchName = &archName[0];

  ioDefPathGet(pPathName);
  if(debug) printf("default Path = %s\n", pathName);
  if(debug) printf("Type-in the architecture (e.g.: mv162 mv167 mv177 pc486 ...):\n");
  gets(pArchName);
  sprintf(pathFrom, "%s/EPICS/%s", pathName, pArchName);
  if(debug) printf("path = %s\n", pathFrom);

  status = chdir(pathFrom);
  if(status != 0) return;

  modId = loadModule((int)"tineLib", LOAD_ALL_SYMBOLS);
  if((int)modId == 0) printf("can't load tineLib !!!\n");
  else if(debug) printf("tineLib module id: %x\n", (int)modId);

  modId = loadModule((int)"e2t.o", LOAD_ALL_SYMBOLS);
  if((int)modId == 0) printf("can't load e2t.o !!!\n");
  else if(debug) printf("e2t.o module id: %x\n", (int)modId);
}

void removeE2TModules()
{
  MODULE_ID modToDel;

  do {
    modToDel = moduleFindByName("tineLib");
    if(modToDel != 0) {
      printf("delete module modToDel = %x\n", (int)modToDel);
      moduleDelete(modToDel);
    }
  } while(modToDel != 0);

  do {
    modToDel = moduleFindByName("e2t.o");
    if(modToDel != 0) {
      printf("delete module modToDel = %x\n", (int)modToDel);
      moduleDelete(modToDel);
    }
  } while(modToDel != 0);

  do {
    modToDel = moduleFindByName("e2tLib");
    if(modToDel != 0) {
      printf("delete module modToDel = %x\n", (int)modToDel);
      moduleDelete(modToDel);
    }
  } while(modToDel != 0);

  printf("All modules for TINE are deleted\n");
}

#endif  /* vxWorks */

void startE2T()
{
int tid;

#ifdef vxWorks
MODULE_ID modToDel;
MODULE_ID modE2TObj;
MODULE_ID modE2TLib;

  modToDel = moduleFindByName("tineLib");
  if(modToDel != 0) {
    modE2TObj = moduleFindByName("e2t.o");
    modE2TLib = moduleFindByName("e2tLib");
    if(modE2TObj != 0 || modE2TLib != 0) {
      tid = taskNameToId("E2T");
      if(tid == ERROR) taskSpawn("E2T", 190, VX_FP_TASK, 10000, (FUNCPTR)E2T,0,0,0,0,0,0,0,0,0,0);
      else printf("Task tid = %d is running, stop or restart possible only !!!\n", tid);
    } else printf("Module <e2t.o> or <e2tLib> failed, load it !!!\n");
  } else printf("Module <tineLib> failed, load it !!!\n");
#endif

#ifdef WIN32
epicsThreadId e2tId;

  e2tId = epicsThreadCreate("E2T", epicsThreadPriorityLow, epicsThreadStackMedium, (EPICSTHREADFUNC)E2T, 0);
  if(!e2tId) printf("error on starting E2T as new thread\n");
  else printf("Thread E2T was started with priority: %d and stack size: %s\n", epicsThreadPriorityLow, "Medium");
#endif  /* WIN32 */
}

#ifdef vxWorks
void stopE2T(int art)
{
int tid, tidHS;
char *pString = "";
int status;
int delTask = FALSE;

  tid = taskNameToId("E2T");
  printf("task E2T has tid = %x\n", tid);

  status = taskStatusString(tid, pString);
  if(status == 0) {
    printf("task E2T %d has status = %s\n", status, pString);
    if((strcmp(pString, "SUSPEND") != 0) ||
       (strcmp(pString, "DELAY+S") != 0) ||
       (strcmp(pString, "PEND+S") != 0) ||
       (strcmp(pString, "PEND+S+T") != 0) ||
       (strcmp(pString, "PEND+S+T+I") != 0) ||
       (strcmp(pString, "DEAD") != 0)) delTask = TRUE;

    if(delTask && art) {
      printf("Close all Sockets, please wait ...\n");
      taskSpawn("haltE2T", 101, VX_FP_TASK, 1000, (FUNCPTR)haltServer,0,0,0,0,0,0,0,0,0,0);
      taskDelay(300);
      tidHS = taskNameToId("haltE2T");
      taskSuspend(tidHS);
      taskDelete(tidHS);
      taskDelete(tid);
      removeE2TModules();
    } else printf("keeping task !!! \n");
  }
}

int restartE2T()
{
int tid, tidHS, i, nr;
char *pString = "";
int status,co=0;
char customInp[8];
char *pCustomInp = &customInp[0];
devChain *dc,*dc1;
char vxW = 1;

  printf("**************************\n");
  printf("Are you sure ? yes/no\n");
  printf("**************************\n\n");
  gets(pCustomInp);
  printf("returned = %s\n", customInp);
  if(strcmp(customInp, "yes") != 0) {
    printf("Cancelled...\n");
    return(-1);
  }

  for(i=0;i<ncomposites;i++) {
    free(composites[i].cmpPrps);
    free(composites[i].cmpDevPrps);
    free(composites[i].cmpDevs);
    free(composites[i].cmpAliasDevs);
    free(composites[i].cmpDesc);
  }
  free(composites);
  printf("numProps = %d\n", numProperties);
  for (nr=0; nr<numProperties; nr++) {
    for (dc=pT[nr].devs,co=0; dc != NULL && !vxMemProbe((char *)dc,VX_READ,1,&vxW) && co < 5; co++)  /* lauft nicht, muss unbedingt besser !!! */
	{
	  printf("vxMemProbe = %d\n", vxMemProbe((char *)dc,VX_READ,1,&vxW));
	  dc1 = dc->next;
	  free(dc);
	  dc = dc1;
	}
  }
  free(pT);
  ncomposites = 0;

  tid = taskNameToId("E2T");
  printf("task E2T has tid = %x\n", tid);

  status = taskStatusString(tid, pString);
  if(status == 0) {
    printf("task E2T %d has status = %s\n", status, pString);
    if(strcmp(pString, "SUSPEND") != 0) {
      taskSuspend(tid);
      printf("task suspended\n");
    } 
    printf("Close all Sockets, please wait ...\n");
    taskSpawn("haltE2T", 101, VX_FP_TASK, 1000, (FUNCPTR)haltServer,0,0,0,0,0,0,0,0,0,0);
    taskDelay(600);
    tidHS = taskNameToId("haltE2T");
    taskSuspend(tidHS);
    taskDelete(tidHS);
    taskDelete(tid);
    taskDelay(300);
    printf("******************************************************\n");
    printf("Reload objects before starting the server !!!\n");
    printf("******************************************************\n\n");
    removeE2TModules();
    /* load first the modules */
    printf("Change to the home/load directory and start startup.tine\n");
  } else {
        printf("Close all Sockets, please wait ...\n");
    taskSpawn("haltE2T", 101, VX_FP_TASK, 1000, (FUNCPTR)haltServer,0,0,0,0,0,0,0,0,0,0);
    taskDelay(600);
    tidHS = taskNameToId("haltE2T");
    taskSuspend(tidHS);
    taskDelete(tidHS);
    taskDelete(tid);
    taskDelay(300);
}
  return(0);
}

#endif  /* vxWorks */
